from random import SystemRandom
from typeid import TypeID
from .conf import settings
from django.contrib.auth import get_user_model
random = SystemRandom()
User = get_user_model()

temp_username_prefix = 'temp'

def create_challenge(length, challenge_characters):
    return "".join(random.choices(challenge_characters, k=length))

def username_generator():
    guid = TypeID()
    response =  {User.USERNAME_FIELD: guid.suffix}
    if settings.UUID_FIELD_NAME:
        response[settings.UUID_FIELD_NAME] = guid.uuid
    return response
    
def token_request_limiter(function):
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)
    return wrapper


def token_redeem_limiter(function):
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapper



from django.utils.timezone import now, timedelta
from django.db.models import Q
from jwt_drf_passwordless.conf import settings
from django.db import transaction, IntegrityError
from .utils import (
    create_challenge,
)
from .models import PasswordlessChallengeToken


class PasswordlessTokenService(object):
    @staticmethod
    def create_token(user, identifier_type):
        # We need to ensure the token is unique, so we'll wrap it in a
        # transaction that retries if the token is not unique.
        tries = 0
        # Only clear tokens that already go beyond the throttle time.
        PasswordlessChallengeToken.objects.delete_expired(
            settings.TOKEN_LIFETIME,
            settings.MAX_TOKEN_USES,
            settings.TOKEN_REQUEST_THROTTLE_SECONDS,
        )
        try:
            with transaction.atomic():
                return PasswordlessTokenService._generate_create_token(
                    user, identifier_type
                )
        except IntegrityError as exception:
            if tries < 5:
                tries += 1
                return PasswordlessTokenService._generate_create_token(
                    user, identifier_type
                )
            else:
                # If we've cannot generate a unique token after 5 tries, we'll
                # raise the exception. Maybe add a message to the admin to cleanup
                # expired stale tokens, or to increase the token length.
                raise exception

    @staticmethod
    def should_throttle(user):
        if settings.TOKEN_REQUEST_THROTTLE_SECONDS:
            return (
                user.jwt_drf_passwordless_tokens.filter(
                    created_at__gt=now() - timedelta(seconds=settings.TOKEN_REQUEST_THROTTLE_SECONDS)
                ).count() > 0
            )
        return False

    @staticmethod
    def _generate_create_token(user, identifier_type):
        # Remove all tokens for this user when issuing a new one
        user.jwt_drf_passwordless_tokens.all().delete()
        token = PasswordlessChallengeToken.objects.create(
            token=create_challenge(
                settings.LONG_TOKEN_LENGTH, settings.LONG_TOKEN_CHARS
            ),
            short_token=create_challenge(
                settings.SHORT_TOKEN_LENGTH, settings.SHORT_TOKEN_CHARS
            ),
            token_request_identifier=identifier_type,
            user=user,
        )
        return token

    @staticmethod
    def check_token(challenge, identifier_field, identifier_value):
        if not challenge:
            return None
        try:
            query = Q(token=challenge)
            if identifier_value and identifier_field:
                query = query | Q(
                    **{
                        "short_token": challenge,
                        "token_request_identifier": identifier_field,
                        "user__" + identifier_field: identifier_value,
                    }
                )
            token = PasswordlessChallengeToken.objects.get(query)
        except PasswordlessChallengeToken.DoesNotExist:
            if (
                identifier_value and identifier_field and settings.INCORRECT_SHORT_TOKEN_REDEEMS_TOKEN
            ):
                # If the token is not found, we'll check if the identifier_value
                # and identifier_field match an existing token. If so, we'll increment the
                # number of attempts for the user. If the user has reached the
                # max number of attempts, we'll lock the user out.
                try:
                    tokens = PasswordlessChallengeToken.objects.filter(
                        **{
                            "token_request_identifier": identifier_field,
                            "user__" + identifier_field: identifier_value,
                        }
                    )
                    for token in tokens:
                        token.redeem()
                except PasswordlessChallengeToken.DoesNotExist:
                    pass

            return None

        if not token.is_valid(settings.TOKEN_LIFETIME, settings.MAX_TOKEN_USES):
            return None

        token.redeem()
        return token



from phonenumber_field.serializerfields import PhoneNumberField
from rest_framework import serializers
from jwt_drf_passwordless.constants import Messages
from django.contrib.auth import get_user_model
from .services import PasswordlessTokenService
from django.contrib.auth.models import update_last_login
from jwt_drf_passwordless.conf import settings
from rest_framework_simplejwt import tokens, settings as jwt_settings

User = get_user_model()


class AbstractPasswordlessTokenRequestSerializer(serializers.Serializer):
    @property
    def token_request_identifier_field(self):
        return NotImplementedError(
            "Passwordless request needs to define at least one field to request a token with."
        )

    def find_user_by_identifier(self, identifier_value):
        try:
            return User.objects.get(
                **{self.token_request_identifier_field: identifier_value}
            )
        except User.DoesNotExist:
            return None

    def validate(self, data):
        validated_data = super().validate(data)
        identifier_value = validated_data[self.token_request_identifier_field]
        user = self.find_user_by_identifier(identifier_value)
        if not settings.REGISTER_NONEXISTENT_USERS and not user:
            raise serializers.ValidationError(Messages.CANNOT_SEND_TOKEN)
        return validated_data

    def create(self, validated_data):
        identifier_value = validated_data[self.token_request_identifier_field]
        user = self.find_user_by_identifier(identifier_value)
        if settings.REGISTER_NONEXISTENT_USERS is True and not user:
            attributes = {
                self.token_request_identifier_field: identifier_value,
                **settings.GENERATORS.username_generator(),
            }
            user = User.objects.create(**attributes)
            if settings.REGISTRATION_SETS_UNUSABLE_PASSWORD:
                user.set_unusable_password()
            user.save()
        return user


class PasswordlessEmailTokenRequestSerializerMixin(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields[settings.EMAIL_FIELD_NAME] = serializers.EmailField(required=True)

    @property
    def token_request_identifier_field(self):
        return settings.EMAIL_FIELD_NAME


class PasswordlessMobileTokenRequestSerializerMixin(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields[settings.MOBILE_FIELD_NAME] = PhoneNumberField(required=True)

    @property
    def token_request_identifier_field(self):
        return settings.MOBILE_FIELD_NAME


class PasswordlessEmailTokenRequestSerializer(
    PasswordlessEmailTokenRequestSerializerMixin,
    AbstractPasswordlessTokenRequestSerializer,
):
    pass


class PasswordlessMobileTokenRequestSerializer(
    PasswordlessMobileTokenRequestSerializerMixin,
    AbstractPasswordlessTokenRequestSerializer,
):
    pass


class PasswordlessJwtRefreshTokenResponse:
    token_class = tokens.RefreshToken

    @classmethod
    def get_token(cls, user):
        return cls.token_class.for_user(user)  # type: ignore

    @classmethod
    def generate_auth_token(cls, user):
        refresh = cls.get_token(user)
        data = {}
        data["refresh"] = str(refresh)
        data["access"] = str(refresh.access_token)
        if jwt_settings.api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, user)
        return data


class AbstractPasswordlessTokenExchangeSerializer(serializers.Serializer):
    token_serializer_class = (
        settings.SERIALIZERS.passwordless_token_response_class
        if settings.SERIALIZERS.passwordless_token_response_class is not None
        else PasswordlessJwtRefreshTokenResponse
    )

    @property
    def token_request_identifier_field(self):
        return None

    default_error_messages = {
        "invalid_credentials": settings.CONSTANTS.messages.INVALID_CREDENTIALS_ERROR
    }
    token = serializers.CharField(required=True)

    @classmethod
    def generate_auth_token(cls, user):
        return cls.token_serializer_class.generate_auth_token(user)  # type: ignore

    def validate(self, attrs):
        super().validate(attrs)
        valid_token = PasswordlessTokenService.check_token(
            attrs.get("token", None),
            self.token_request_identifier_field,
            attrs.get(self.token_request_identifier_field, None),
        )
        if valid_token:
            self.user = valid_token.user
            return self.generate_auth_token(self.user)
        self.fail("invalid_credentials")


class PasswordlessEmailTokenExchangeSerializer(
    PasswordlessEmailTokenRequestSerializerMixin,
    AbstractPasswordlessTokenExchangeSerializer,
):
    pass


class PasswordlessMobileTokenExchangeSerializer(
    PasswordlessMobileTokenRequestSerializerMixin,
    AbstractPasswordlessTokenExchangeSerializer,
):
    pass



from django.conf import settings as django_settings
from django.test.signals import setting_changed
from django.utils.functional import LazyObject
from django.utils.module_loading import import_string

JWT_DRF_PASSWORDLESS_SETTINGS_NAMESPACE = "JWT_DRF_PASSWORDLESS"


class ObjDict(dict):
    def __getattribute__(self, item):
        try:
            val = self[item]
            if isinstance(val, str):
                val = import_string(val)
            elif isinstance(val, (list, tuple)):
                val = [import_string(v) if isinstance(v, str) else v for v in val]
            self[item] = val
        except KeyError:
            val = super().__getattribute__(item)
        return val


default_settings = {
    "SHORT_TOKEN_LENGTH": 6,
    "LONG_TOKEN_LENGTH": 64,
    "SHORT_TOKEN_CHARS": "0123456789",
    "LONG_TOKEN_CHARS": "abcdefghijklmnopqrstuvwxyz0123456789",
    "TOKEN_LIFETIME": 600,
    "TOKEN_REQUEST_THROTTLE_SECONDS": 60,
    "ALLOW_ADMIN_AUTHENTICATION": False,
    "REGISTER_NONEXISTENT_USERS": False,
    "REGISTRATION_SETS_UNUSABLE_PASSWORD": True,
    "EMAIL_FIELD_NAME": "email",
    "MOBILE_FIELD_NAME": "phone_number",
    "UUID_FIELD_NAME": None,
    # If true, an attempt to redeem a token with the wrong token type
    # will count for the times a token has been used
    "INCORRECT_SHORT_TOKEN_REDEEMS_TOKEN": False,
    "ALLOWED_PASSWORDLESS_METHODS": ["EMAIL"],  # ["EMAIL", "MOBILE"]
    "PASSWORDLESS_EMAIL_LOGIN_URL": None,
    "MAX_TOKEN_USES": 1,
    "GENERATORS": ObjDict(
        {
            "username_generator": "jwt_drf_passwordless.utils.username_generator",
        }
    ),
    "EMAIL": ObjDict(
        {
            "passwordless_request": "jwt_drf_passwordless.email.PasswordlessRequestEmail",
        }
    ),
    "SERIALIZERS": ObjDict(
        {
            "passwordless_token_response_class": None,
            "passwordless_email_token_request": "jwt_drf_passwordless.serializers.PasswordlessEmailTokenRequestSerializer",
            "passwordless_mobile_token_request": "jwt_drf_passwordless.serializers.PasswordlessMobileTokenRequestSerializer",
            "passwordless_email_token_exchange": "jwt_drf_passwordless.serializers.PasswordlessEmailTokenExchangeSerializer",
            "passwordless_mobile_token_exchange": "jwt_drf_passwordless.serializers.PasswordlessMobileTokenExchangeSerializer",
        }
    ),
    "PERMISSIONS": ObjDict(
        {
            "passwordless_token_exchange": ["rest_framework.permissions.AllowAny"],
            "passwordless_token_request": ["rest_framework.permissions.AllowAny"],
        }
    ),
    "DECORATORS": ObjDict(
        {
            "token_request_rate_limit_decorator": "jwt_drf_passwordless.utils.token_request_limiter",
            "token_redeem_rate_limit_decorator": "jwt_drf_passwordless.utils.token_redeem_limiter",
        }
    ),
    "SMS_SENDERS": ObjDict(
        {
            "passwordless_request": "jwt_drf_passwordless.sms.PasswordlessRequestSMS",
        }
    ),
    "CONSTANTS": ObjDict({"messages": "jwt_drf_passwordless.constants.Messages"}),
}


SETTINGS_TO_IMPORT = []


class Settings:
    def __init__(self, default_settings, explicit_overriden_settings: dict = None):
        if explicit_overriden_settings is None:
            explicit_overriden_settings = {}

        overriden_settings = (
            getattr(django_settings, JWT_DRF_PASSWORDLESS_SETTINGS_NAMESPACE, {})
            or explicit_overriden_settings
        )

        self._load_default_settings()
        self._override_settings(overriden_settings)
        self._init_settings_to_import()

    def _load_default_settings(self):
        for setting_name, setting_value in default_settings.items():
            if setting_name.isupper():
                setattr(self, setting_name, setting_value)

    def _override_settings(self, overriden_settings: dict):
        for setting_name, setting_value in overriden_settings.items():
            value = setting_value
            if isinstance(setting_value, dict):
                value = getattr(self, setting_name, {})
                value.update(ObjDict(setting_value))
            setattr(self, setting_name, value)

    def _init_settings_to_import(self):
        for setting_name in SETTINGS_TO_IMPORT:
            value = getattr(self, setting_name)
            if isinstance(value, str):
                setattr(self, setting_name, import_string(value))


class LazySettings(LazyObject):
    def _setup(self, explicit_overriden_settings=None):
        self._wrapped = Settings(default_settings, explicit_overriden_settings)


settings = LazySettings()


def reload_settings(*args, **kwargs):
    global settings
    setting, value = kwargs["setting"], kwargs["value"]
    if setting == JWT_DRF_PASSWORDLESS_SETTINGS_NAMESPACE:
        settings._setup(explicit_overriden_settings=value)


setting_changed.connect(reload_settings)



from django.urls import re_path

from . import views

urlpatterns = [
    re_path(
        r"^request/email/$",
        views.PasswordlessEmailTokenRequestView.as_view(),
        name="passwordless_email_signup_request",
    ),
    re_path(
        r"^request/mobile/$",
        views.PasswordlessMobileTokenRequestView.as_view(),
        name="passwordless_mobile_signup_request",
    ),
    re_path(
        r"^exchange/mobile/$",
        views.MobileExchangePasswordlessTokenForAuthTokenView.as_view(),
        name="mobile_passwordless_token_exchange",
    ),
    re_path(
        r"^exchange/email/$",
        views.EmailExchangePasswordlessTokenForAuthTokenView.as_view(),
        name="email_passwordless_token_exchange",
    )
]



# Generated by Django 5.0.3 on 2024-03-06 00:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jwt_drf_passwordless", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="passwordlesschallengetoken",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="passwordlesschallengetoken",
            name="short_token",
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name="passwordlesschallengetoken",
            name="token",
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name="passwordlesschallengetoken",
            name="token_request_identifier",
            field=models.CharField(max_length=255),
        ),
    ]



from templated_mail.mail import BaseEmailMessage

from jwt_drf_passwordless.conf import settings


class PasswordlessRequestEmail(BaseEmailMessage):
    template_name = "email/passwordless_request.html"

    def get_context_data(self):
        context = super().get_context_data()
        if settings.PASSWORDLESS_EMAIL_LOGIN_URL:
            # Eg magic links / Deep links for mobile apps
            context["url"] = settings.PASSWORDLESS_EMAIL_LOGIN_URL.format(**context)
        return context



from django.db import models
from django.conf import settings
from django.utils.timezone import now
from datetime import timedelta


class PasswordlessChallengeTokenManager(models.Manager):
    def delete_expired(
        self, token_lifetime_seconds, max_token_uses, only_older_than: None
    ):
        query = models.Q(
            created_at__lt=now() - timedelta(seconds=token_lifetime_seconds)
        ) | models.Q(uses__gte=max_token_uses)

        if only_older_than:
            query = query | models.Q(
                created_at__lt=now() - timedelta(seconds=only_older_than)
            )

        return self.filter(query).delete()


class PasswordlessChallengeToken(models.Model):
    objects = PasswordlessChallengeTokenManager()

    # We will deliver two tokens. One which is long and one which is short.
    # The short one needs to be redeemed with the same identifier that it was
    # sent to. This is a mitigation to brute force attacks, that might be able
    # to break a short token.
    # The long token can be redeemed without any other information, as it is
    # significantly harder to brute force.

    token = models.CharField(unique=True, max_length=255)
    short_token = models.CharField(max_length=255)
    created_at = models.DateTimeField(auto_now=True)
    uses = models.IntegerField(default=0)
    token_request_identifier = models.CharField(max_length=255)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="jwt_drf_passwordless_tokens",
        null=True,
        on_delete=models.CASCADE,
    )

    def redeem(self):
        self.uses += 1
        return self.save()

    def is_valid(self, token_lifetime_seconds, max_token_uses):
        if self.created_at + timedelta(seconds=token_lifetime_seconds) < now():
            return False
        if self.uses >= max_token_uses:
            return False
        return True



# Generated by Django 3.2.18 on 2023-04-23 07:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PasswordlessChallengeToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.TextField(unique=True)),
                ('short_token', models.TextField()),
                ('created_at', models.DateTimeField(auto_now=True)),
                ('uses', models.IntegerField(default=0)),
                ('token_request_identifier', models.TextField()),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='jwt_drf_passwordless_tokens', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]
